# ---------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# ---------------------------------------------------------
"""Add columns indicating corresponding numeric columns have NaN."""
from typing import Optional, List
import logging

import numpy
import pandas as pd
import warnings

from azureml.automl.core.shared.logging_utilities import function_debug_log_wrapped
from ..automltransformer import AutoMLTransformer
from azureml.automl.runtime.shared.time_series_data_frame import TimeSeriesDataFrame


# Prevent warnings when using Jupyter
warnings.simplefilter("ignore")
pd.options.mode.chained_assignment = None


class MissingDummiesTransformer(AutoMLTransformer):
    """Add columns indicating corresponding numeric columns have NaN."""

    MARKER_VALUE_MISSING = 1
    MARKER_VALUE_NOT_MISSING = 0

    MISSING_DUMMIES_POSTFIX = '_WASNULL'

    def __init__(self, numerical_columns: List[str]) -> None:
        """
        Construct for MissingDummiesTransformer.

        :param numerical_columns: The columns that will be marked.
        :type numerical_columns: list
        :return:
        """
        super().__init__()
        self.numerical_columns = numerical_columns

    @function_debug_log_wrapped(logging.INFO)
    def fit(self, x: TimeSeriesDataFrame, y: Optional[numpy.ndarray] = None) -> 'MissingDummiesTransformer':
        """
        Fit function for MissingDummiesTransformer.

        :param x: Input data.
        :type x: azureml.automl.runtime.shared.time_series_data_frame.TimeSeriesDataFrame
        :param y: Target values.
        :type y: numpy.ndarray
        :return: Class object itself.
        """
        return self

    @function_debug_log_wrapped(logging.INFO)
    def transform(self, x: TimeSeriesDataFrame) -> TimeSeriesDataFrame:
        """
        Transform function for MissingDummiesTransformer.

        :param x: Input data.
        :type x: azureml.automl.runtime.shared.time_series_data_frame.TimeSeriesDataFrame
        :return: Result of MissingDummiesTransformer.
        """
        for col in self.numerical_columns:
            x[MissingDummiesTransformer.get_column_name(col)] = x[col].isnull().astype('int')
        return x

    @staticmethod
    def get_column_name(col: str) -> str:
        """
        Return the name of column marking nan in the given source column.

        :param col: the name of the source column.
        :return: the name of new feature column.
        """
        return col + MissingDummiesTransformer.MISSING_DUMMIES_POSTFIX

    @staticmethod
    def get_col_internal_type(col_name: str) -> Optional[str]:
        """
        Get the type of a column if it is generated by missing dummies transformer.

        :param col_name: The column name.
        :return: If column is generated by missing dummies transformer, return 'missing_dummies', else None.
        """
        if col_name.endswith(MissingDummiesTransformer.MISSING_DUMMIES_POSTFIX):
            return "missing_dummies"
        return None
